/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * Loads a shared object library or dynamic link library and provides
 * it with an implementation of a virtual POSIX-like interface that
 * sits directly on top of the underlying POSIX-like system.
 * Alternatively provides the trivial embedding thereof.
 *
 * If the environment variable "VRPSX_LOADER_TRACE_FILE" is defined
 * and has a nonempty value, then that value is expanded to produce
 * a pathname (see the function header comments for "trace_expand"
 * in this file) and this program writes interface call traces to
 * the named file.  As yet, not all interface calls are traced.
 */

#define _POSIX_C_SOURCE 200809L
#define _XOPEN_SOURCE   500

#include "implementation.h"
#include "support_map.h"
#include <vrpsx_identified_function.h>
#include <vrpsx_identified_function_identifier.h>
#include <vrpsx_system_initial_identifier.h>
#include <vrpsx_system_main.h>
#include <vrpsx_system_main_identifier.h>
#include <vrpsx_embedding_initial_identifier.h>
#include <vrpsx_embedding_main.h>
#include <vrpsx_embedding_main_identifier.h>
#include <vrpsx_utility.h>
#include <sys/utsname.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <assert.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <ctype.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <dlfcn.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <errno.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <iconv.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <langinfo.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <limits.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <locale.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <stdio.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <stdlib.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <string.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <time.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <unistd.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <wchar.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */

#if defined(__CYGWIN__)
#include <cygwin/version.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <sys/cygwin.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#include <w32api/windows.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#undef interface /* we want to use this as an identifier */
#endif

extern char **environ;


#include <loader_support_map.c>


#if defined(__linux__) || defined(__CYGWIN__)

/* Symbolic link to executable pathname.
 */
static char const proc_self_exe[] = "/proc/self/exe";

#endif

#if defined(__CYGWIN__) && CYGWIN_VERSION_DLL_MAJOR >= 1007

/* Environment variable used to limit the self-exec behavior under Cygwin 1.7+.
 *
 * We chose part of the name of this environment variable randomly
 * in order to minimize the chance of accidental collision with
 * a user-chosen variable, which could cause erasure of that variable.
 *
 * This variable should not be seen through the interface or by child
 * processes once it is pruned back out.
 */
static char const VRPSX_LOADER_RECURSIVE[] = "VRPSX_LOADER_RECURSIVE_d1c4b3ff";

#endif /* defined(__CYGWIN__) && CYGWIN_VERSION_DLL_MAJOR >= 1007 */


extern vrpsx_interface_identifier_t const *caller_supports(
    function_binding_t const support_map[],
    size_t dimension,
    vrpsx_interface_identifier_t interface_identifier)
{
    function_binding_t const *binding
            = (function_binding_t const *) bsearch(
                &interface_identifier,
                support_map,
                dimension,
                sizeof support_map[0],
                &direct_interface_identifier_compare);

    if (binding) {
        return &binding->provided;
    } else {
        return NULL;
    }
}


static void *malloc_or_exit(size_t size)
{
    void *memory = malloc(size);
    if (! memory) {
        perror("malloc");
        exit(1);
    }
    return memory;
}


typedef struct operations_end_struct {
} operations_end_t;

typedef struct operations_struct {
    char *(*get_library_path)(
        char const *argv0);

    void *(*dlopen)(
        char const *path,
        int flags);

    void *(*dlsym)(
        void *library,
        char const *identifier);

    void (*synchronize_environment)(void);

    /* End marker for documentation and to help detect initialization bugs.
     */
    operations_end_t const *operations_end;
} operations_t;

static operations_end_t const operations_end;

static operations_t const *op;


#if defined(__CYGWIN__)

/* Tells Cygwin to bring the Windows environment up to date.
 */
static void synchronize_environment(void)
{
    /* This call is documented by "Restricted Win32 environment" in the
     * "Cygwin User's Guide" (at least in its Cygwin 1.7.16 variant).
     */
    cygwin_internal(CW_SYNC_WINENV);
}

static void trace_synchronize_environment(void)
{
    static char const me[] = "cygwin_internal";
    trace_stack_t top;
    uintptr_t result;

    trace_enter_begin(&top, me);
    trace_identifier(&top, "CW_SYNC_WINENV");
    trace_end(&top);

    result = cygwin_internal(CW_SYNC_WINENV);

    trace_leave_begin(&top, me);
    trace_uintptr(&top, result);
    trace_end(&top);
}

#else

static void synchronize_environment(void) {}

static void trace_synchronize_environment(void) {}

#endif


static void static_argv_release(
    vrpsx_argv_t **array)
{}

static void static_argv_acquire(
    vrpsx_argv_t *array)
{}

static vrpsx_reference_counted_end_t const reference_counted_end;

static vrpsx_argv_end_t const argv_end;

static vrpsx_argv_vtbl_t const static_argv_vtbl = {
    &static_argv_release,
    &static_argv_acquire,
    &reference_counted_end,
    &argv_end
};

static vrpsx_argv_t arguments = {
    &static_argv_vtbl,
    NULL,
    0
};


typedef struct dynamic_argv_struct {
    vrpsx_argv_t interface;
    vrpsx_reference_count_t reference_count;
    char *array[1]; /* the actual dimension is computed at run time */
    /* the text pointed to by elements of "array" comes after that array */
} dynamic_argv_t;

static void dynamic_argv_release(
    vrpsx_argv_t **array)
{
    VRPSX_UPCAST_NONNULL_DECL(dynamic_argv_t, , obj, interface, *array);
    vrpsx_reference_count_t new_count;

    *array = NULL;

    VRPSX_REFERENCE_COUNT_DECR_READ(&obj->reference_count, &new_count);

    if (new_count == 0) {
        free(obj);
    }
}

static void dynamic_argv_acquire(
    vrpsx_argv_t *array)
{
    VRPSX_UPCAST_NONNULL_DECL(dynamic_argv_t, , obj, interface, array);
    VRPSX_REFERENCE_COUNT_INCR(&obj->reference_count);
}

static vrpsx_argv_vtbl_t const dynamic_argv_vtbl = {
    &dynamic_argv_release,
    &dynamic_argv_acquire,
    &reference_counted_end,
    &argv_end
};

static void dynamic_argv_copy_inner_environment(
    vrpsx_argv_t **out_array)
{
    int array_dimension;
    size_t object_size, string_size;
    dynamic_argv_t *obj;
    char **ep, **ap;
    char *assign;

    *out_array = NULL;

    array_dimension = 1; /* for NULL termination */
    for (ep = environ; *ep; ++ep) {
        ++ array_dimension;
    }

    object_size = offsetof(dynamic_argv_t, array)
            + (array_dimension * sizeof(char *));
    for (ep = environ; *ep; ++ep) {
        object_size += strlen(*ep) + 1;
    }

    obj = (dynamic_argv_t *)malloc_or_exit(object_size);
    obj->interface.vtbl = &dynamic_argv_vtbl;

    ap = obj->array;
    VRPSX_ARGV_INITIALIZE_ARRAY(&obj->interface, array_dimension - 1, ap);

    obj->reference_count = 1;

    assign = (char *)(ap + array_dimension);
    for (ep = environ; *ep; ++ep) {
        string_size = strlen(*ep) + 1;
        memcpy(assign, *ep, string_size);
        *ap++ = assign;
        assign += string_size;
    }

    *ap = NULL;

    assert ((char *)(ap + 1) == obj->array[0]);
    assert ((char *)obj + object_size == assign);

    *out_array = &obj->interface;
}

#if defined(__CYGWIN__)

static void dynamic_argv_copy_outer_environment(
    vrpsx_argv_t **out_array)
{
    wchar_t *wp, *wide_concat_begin;
    size_t wide_concat_words;
    dynamic_argv_t *obj;
    char *np, *narrow_concat_end;
    size_t narrow_concat_bytes, total_bytes;
    char **ap;
    int array_dimension;
    iconv_t cd;
    char
#if CYGWIN_VERSION_DLL_MAJOR == 1005
            const
#endif
            *in_ptr;
    size_t in_rem;
    char *out_ptr;
    size_t out_rem;

    *out_array = NULL;

    /* Tell Cygwin to bring the Windows environment up to date.
     *
     * This step is necessary not only to see changes made earlier
     * in this process, but also just to get the full environment
     * in the case where the parent process is a Cygwin process.
     */
    op->synchronize_environment();

    wide_concat_begin = GetEnvironmentStringsW();
    if (! wide_concat_begin) {
        /* Perhaps some day we will go through the lengthy procedure for
         * formatting a Windows error.  But failure is rather unlikely.
         */
        fputs("GetEnvironmentStringsW failed\n", stderr);
        exit(1);
    }

    array_dimension = 1; /* for NULL termination */
    for (wp = wide_concat_begin; *wp; wp += wcslen(wp) + 1) {
        ++ array_dimension;
    }
    wide_concat_words = wp - wide_concat_begin;

    /* Since 2003, UTF-8 encodings max out at 4 octets.
     */
    narrow_concat_bytes = 4 * wide_concat_words;
    total_bytes = offsetof(dynamic_argv_t, array)
            + (array_dimension * sizeof(char *))
            + narrow_concat_bytes;
    obj = (dynamic_argv_t *)malloc_or_exit(total_bytes);
    obj->interface.vtbl = &dynamic_argv_vtbl;
    obj->reference_count = 1;

    cd = iconv_open("UTF-8", "UTF-16LE");
    if (cd == (iconv_t)(-1)) {
        perror("iconv_open");
        exit(1);
    }

    in_ptr = (char *)wide_concat_begin;
    in_rem = wide_concat_words * sizeof(wchar_t);
    out_ptr = (char *)(obj->array + array_dimension);
    out_rem = narrow_concat_bytes;
    if (iconv(cd, &in_ptr, &in_rem, &out_ptr, &out_rem) == (size_t)(-1)) {
        perror("iconv");
        exit(1);
    }

    if (iconv_close(cd)) {
        perror("iconv_close");
        exit(1);
    }

    if (! FreeEnvironmentStringsW(wide_concat_begin)) {
        /* Perhaps some day we will go through the lengthy procedure for
         * formatting a Windows error.  But failure is highly unlikely.
         */
        fputs("FreeEnvironmentStringsW failed\n", stderr);
        exit(1);
    }

    total_bytes -= out_rem;
    obj = (dynamic_argv_t *)realloc(obj, total_bytes);
    if (! obj) {
        perror("realloc");
        exit(1);
    }

    ap = obj->array;
    VRPSX_ARGV_INITIALIZE_ARRAY(&obj->interface, array_dimension - 1, ap);

    narrow_concat_end = (char *)obj + total_bytes;
    for (np = (char *)(ap + array_dimension);
         np < narrow_concat_end; np += strlen(np) + 1) {
        *ap++ = np;
    }
    *ap = NULL;

    assert ((char *)(ap + 1) == obj->array[0]);
    assert ((char *)obj + total_bytes == np);

    *out_array = &obj->interface;
}

#endif /* defined(__CYGWIN__) */


static void loader_system_initial_replace_interface(
    vrpsx_system_initial_t **identified,
    vrpsx_interface_identifier_t interface_identifier,
    void **out_alternative,
    vrpsx_error_t **out_error)
{
    size_t offset;

    *out_alternative = NULL;
    *out_error = NULL;

    vrpsx_interface_identifier_t const *provided
            = loader_system_initial_supports(interface_identifier, &offset);

    if (! provided) {
        direct_error_create(ENOSYS, out_error);
        return;
    }

    switch (*provided) {
        case vrpsx_system_initial_IDENTIFIER: {
            *out_alternative = (char *)*identified + offset;
            *identified = NULL;
            return;
        }
    }

    assert(0 /* provided interface needs case in above switch */);
    direct_error_create(ENOSYS, out_error);
}

static vrpsx_identified_end_t const identified_end;

static vrpsx_system_initial_end_t const system_initial_end;

static vrpsx_system_initial_vtbl_t const loader_system_initial_vtbl = {
    vrpsx_system_initial_IDENTIFIER,
    &loader_system_initial_replace_interface,
    &identified_end,
    &system_initial_end
};


static void loader_embedding_initial_replace_interface(
    vrpsx_embedding_initial_t **identified,
    vrpsx_interface_identifier_t interface_identifier,
    void **out_alternative,
    vrpsx_error_t **out_error)
{
    size_t offset;

    *out_alternative = NULL;
    *out_error = NULL;

    vrpsx_interface_identifier_t const *provided
            = loader_embedding_initial_supports(interface_identifier, &offset);

    if (! provided) {
        direct_error_create(ENOSYS, out_error);
        return;
    }

    switch (*provided) {
        case vrpsx_embedding_initial_IDENTIFIER: {
            *out_alternative = (char *)*identified + offset;
            *identified = NULL;
            return;
        }
    }

    assert(0 /* provided interface needs case in above switch */);
    direct_error_create(ENOSYS, out_error);
}

static vrpsx_embedding_initial_end_t const embedding_initial_end;

static vrpsx_embedding_initial_vtbl_t const loader_embedding_initial_vtbl = {
    vrpsx_embedding_initial_IDENTIFIER,
    &loader_embedding_initial_replace_interface,
    &identified_end,
    &system_initial_end,
    &embedding_initial_end
};


typedef struct system_preparation_struct {
    vrpsx_system_main_t *function;
    vrpsx_system_initial_t initial;
} system_preparation_t;


typedef struct embedding_preparation_struct {
    vrpsx_embedding_main_t *function;
    vrpsx_embedding_initial_t initial;
} embedding_preparation_t;


/* The union of the various preparations that we might perform in
 * order to call an identified function that we dynamically loaded.
 *
 * Here we also declare the singleton instance of this union.
 */
static union preparation_union {
    system_preparation_t system;
    embedding_preparation_t embedding;
} preparation;


/* A function that uses the (preparation) singleton to call
 * some identified function and return its result.  Assumes
 * that that singleton was initialized in a compatible way.
 */
typedef int invoker(void);


static int system_preparation_call(void)
{
    return preparation.system.function(
        &preparation.system.initial);
}

/* This constructor expects that you already filled in argc, argv, and env.
 */
static invoker *system_preparation_construct(
    void (*function)(void),
    vrpsx_embedding_t **embedding)
{
    system_preparation_t *prep = &preparation.system;
    vrpsx_system_initial_t *initial = &prep->initial;

    prep->function = (vrpsx_system_main_t*)function;

    initial->vtbl = &loader_system_initial_vtbl;

    initial->arguments = &arguments;

    dynamic_argv_copy_inner_environment(&initial->environment);

    VRPSX_COUNTED_REFERENCE_ACQUIRE(
        vrpsx_system_t,
        &initial->system,
        &(*embedding)->system);
    VRPSX_COUNTED_REFERENCE_CLEAR(vrpsx_embedding_t, embedding);

    return &system_preparation_call;
}


static int embedding_preparation_call(void)
{
    return preparation.embedding.function(
        &preparation.embedding.initial);
}

/* This constructor expects that you already filled in argc, argv, and env.
 */
static invoker *embedding_preparation_construct(
    void (*function)(void),
    vrpsx_embedding_t **embedding)
{
    embedding_preparation_t *prep = &preparation.embedding;
    vrpsx_embedding_initial_t *initial = &prep->initial;

    prep->function = (vrpsx_embedding_main_t*)function;

    initial->vtbl = &loader_embedding_initial_vtbl;

    initial->arguments = &arguments;

    dynamic_argv_copy_inner_environment(&initial->environment);

    VRPSX_COUNTED_REFERENCE_ACQUIRE(
        vrpsx_system_t,
        &initial->system,
        &(*embedding)->system);

#if defined(__CYGWIN__)
    dynamic_argv_copy_outer_environment(&initial->outer_environment);
#else
    VRPSX_COUNTED_REFERENCE_ACQUIRE(
        vrpsx_argv_t,
        &initial->outer_environment,
        initial->environment);
#endif

    VRPSX_COUNTED_REFERENCE_MOVE(
        vrpsx_embedding_t,
        &initial->embedding,
        embedding);

    return &embedding_preparation_call;
}


/* The library path is the name of this program, resolved to
 * a particular directory, stripped of any executable suffix,
 * with the dynamic library suffix appended.
 */
static char *get_library_path(
    char const *argv0)
{
    char *resolution = NULL;
    size_t length;
    size_t ext_length;
    char const *ext;
    char *path;

    /* Find directory containing executable by using /proc/self/exe,
     * which is more accurate than searching PATH, but less portable.
     */
#if defined(__linux__) || defined(__CYGWIN__)
    {
        char target[PATH_MAX];
        ssize_t target_length;

        target_length = readlink(proc_self_exe, target, sizeof target);
        if (target_length == -1) {
            int ern = errno;
            fprintf(stderr, "%s: %s: %s\n",
                    argv0, proc_self_exe, strerror(ern));
            return NULL;
        }
        if (target_length >= sizeof target) {
            fprintf(stderr, "%s: %s: link target too long\n",
                    argv0, proc_self_exe);
            return NULL;
        }

        resolution = (char *)malloc_or_exit(target_length + 1);
        memcpy(resolution, target, target_length);
        resolution[target_length] = '\0';
        argv0 = resolution;
    }
#else

#error not yet implemented: search PATH for argv[0]

#endif

    length = strlen(argv0);

#if defined(__CYGWIN__)

    if (length >= 4) {
        char const *tail = argv0 + length - 4;
        if (tail[0] == '.'
                && (tail[1] == 'e' || tail[1] == 'E')
                && (tail[2] == 'x' || tail[2] == 'X')
                && (tail[3] == 'e' || tail[3] == 'E')) {
            length -= 4;
        }
    }

    ext = ".dll";
    ext_length = 4;

#else

    ext_length = 3;
    ext = ".so";

#endif

    path = (char *)malloc_or_exit(length + ext_length + 1);

    memcpy(path, argv0, length);
    memcpy(path + length, ext, ext_length);
    path[length + ext_length] = '\0';
    free(resolution);
    return path;
}

static char *trace_get_library_path(
    char const *argv0)
{
    static char const me[] = "get_library_path";
    trace_stack_t top;
    char *library_path;

    trace_enter_begin(&top, me);
    trace_string(&top, argv0);
    trace_end(&top);

    library_path = get_library_path(argv0);

    trace_leave_begin(&top, me);
    trace_string(&top, library_path);
    trace_end(&top);

    return library_path;
}


static void *trace_dlopen(
    char const *path,
    int flags)
{
    static char const me[] = "dlopen";
    static flag_definition_t const flagdefs[] = {
        { (RTLD_LAZY | RTLD_NOW),     RTLD_LAZY,   "RTLD_LAZY"   },
        { (RTLD_LAZY | RTLD_NOW),     RTLD_NOW,    "RTLD_NOW"    },
#if defined(RTLD_LOCAL)
        { (RTLD_GLOBAL | RTLD_LOCAL), RTLD_GLOBAL, "RTLD_GLOBAL" },
        { (RTLD_GLOBAL | RTLD_LOCAL), RTLD_LOCAL,  "RTLD_LOCAL", },
#else
        { RTLD_GLOBAL,                RTLD_GLOBAL, "RTLD_GLOBAL" },
#endif
        { 0, 0, NULL }
    };
    trace_stack_t top;
    void *library;

    trace_enter_begin(&top, me);
    trace_string(&top, path);
    trace_flagset(&top, flagdefs, flags);
    trace_end(&top);

    library = dlopen(path, flags);

    trace_leave_begin(&top, me);
    trace_address(&top, library);
    trace_end(&top);

    return library;
}


static void *trace_dlsym(
    void *library,
    char const *identifier)
{
    static char const me[] = "dlsym";
    trace_stack_t top;
    void *address;

    trace_enter_begin(&top, me);
    trace_address(&top, library);
    trace_string(&top, identifier);
    trace_end(&top);

    address = dlsym(library, identifier);

    trace_leave_begin(&top, me);
    trace_address(&top, address);
    trace_end(&top);

    return address;
}


/* The interfaces that we support mandate UTF-8 encoding of strings,
 * but that we otherwise set the locale based upon the environment.
 *
 * This function accomplishes that locale choice--except in the Cygwin 1.5
 * case where we have not yet found a way to ensure UTF-8 encoding.  Please
 * feel free to fix that bug.
 */
static void establish_locale(
    char const *const *argv)
{
#if ! ( defined(__CYGWIN__) && CYGWIN_VERSION_DLL_MAJOR < 1007 )
    /* Internationalization variables that may mention the character set
     * and affect the automatic choice of locale.  We may need to adjust
     * these if they do not already specify the UTF-8 encoding we desire.
     *
     * Each locale variable is prefixed by '1' if it is one of the locale
     * variables that we think are essential to setting the encoding, and
     * to '0' otherwise.
     */
    static char const *const locale_variables[] = {
        "0LC_COLLATE",
        "0LC_MESSAGES",
        "0LC_MONETARY",
        "0LC_NUMERIC",
        "0LC_TIME"
        "1LANG",
        "1LC_ALL",
        "1LC_CTYPE",
    };

    char const *const *locale_variable;
    char const *name, *value, *codeset, *modifier, *end;
    char *new_value, *next;
    size_t prefix, add_dot, suffix;
    int essential;
#endif

#if defined(__CYGWIN__) && CYGWIN_VERSION_DLL_MAJOR >= 1007
    char target[PATH_MAX];
    ssize_t target_length;
#endif

    if (! setlocale(LC_ALL, "")) {
        /* The locale settings in the environment could not be satisfied.
         * In this case we override the default in order to ensure UTF-8.
         *
         * In this case Cygwin 1.7 process startup code would have already
         * chosen UTF-8 for Windows pathname translation, and so we do not
         * need to restart the process in this particular case.
         */
        setlocale(LC_ALL, "C.UTF-8");
        goto unset;
    }

#if defined(__CYGWIN__) && CYGWIN_VERSION_DLL_MAJOR < 1007

    /* In Cygwin 1.5 we have not yet figured out how to ensure UTF-8.
     */
    return;

#else /* Not Cygwin 1.5 */

    if (0 == strcmp(nl_langinfo(CODESET), "UTF-8")) {
        /* We are using UTF-8 already, though under Cygwin 1.7+
         * that may be because we re-exec-ed this process.
         */
        goto unset;
    }

#if defined(__CYGWIN__) && CYGWIN_VERSION_DLL_MAJOR >= 1007

    /* Did we already try to re-exec and fail?  If so then do not try again.
     */
    if (getenv(VRPSX_LOADER_RECURSIVE)) {
        fprintf(stderr, "%s: WARNING: failed to switch to UTF-8\n", argv[0]);
        goto unset;
    }
    if (setenv(VRPSX_LOADER_RECURSIVE, "1", 1)) {
        perror("setenv");
        exit(1);
    }

#endif /* defined(__CYGWIN__) && CYGWIN_VERSION_DLL_MAJOR >= 1007 */

    /* Adjust the locale environment variables to specify UTF-8.
     */
    essential = 0;
    for (locale_variable = locale_variables;
         locale_variable != VRPSX_ARRAY_END(locale_variables);
         ++ locale_variable) {
        name = *locale_variable + 1;
        value = getenv(name);
        if (! value) {
            continue;
        }

        /* Have we found one of the variables we think controls the encoding?
         * But regardless, we still edit the encoding field for this variable.
         */
        if (name[-1] == '1') {
            essential = 1;
        }

        /* Find the beginning of the codeset specifier,
         * or where it would go if it were present.  We will
         * need to add a '.' if there is no codeset specifier.
         */
        add_dot = 1;
        for (codeset = value; *codeset && *codeset != '@';) {
            if (*codeset++ == '.') {
                add_dot = 0;
                break;
            }
        }

        /* Find the end of the codeset specifier,
         * or where it would go if it were present.
         */
        for (modifier = codeset; *modifier && *modifier != '@'; ++modifier) {}

        /* Find the end of the environment variable value.
         */
        for (end = modifier; *end; ++end) {}

        if (modifier - codeset == 5 && ! memcmp(codeset, "UTF-8", 5)) {
            /* No change required for this particular variable.
             */
            continue;
        }

        /* The length of the part before the codeset.
         */
        prefix = codeset - value;

        /* The length of the part after the codeset, including NUL termination.
         */
        suffix = end + 1 - modifier;

        /* Allocate space for a temporary copy of the new value.
         */
        new_value = (char *)malloc_or_exit(prefix + add_dot + 5 + suffix);

        /* Copy everything over except the codeset, which becomes "UTF-8".
         */
        next = new_value;
        MEMCPY_END(next, value, prefix);
        MEMCPY_END(next, ".", add_dot);
        MEMCPY_END(next, "UTF-8", 5);
        memcpy(next, modifier, suffix);

        /* Replace the old environment variable assignment.
         */
        if (setenv(name, new_value, 1)) {
            perror("setenv");
            exit(1);
        }

        /* Free the temporary copy of the new value.
         */
        free(new_value);
    }

    if (! essential) {
        /* We may need to override the default character encoding by
         * defining a locale environment variable not previously defined.
         * We pick the one with the most narrow impact: LC_CTYPE.
         */
        if (setenv("LC_CTYPE", "C.UTF-8", 0)) {
            perror("setenv");
            exit(1);
        }
    }

#if defined(__CYGWIN__) && CYGWIN_VERSION_DLL_MAJOR >= 1007

    target_length = readlink(proc_self_exe, target, sizeof target);
    if (target_length == -1) {
        perror(proc_self_exe);
        exit(1);
    }
    if (target_length >= sizeof target) {
        errno = ENAMETOOLONG;
        perror(proc_self_exe);
        exit(1);
    }
    target[target_length] = '\0';

    /* Cygwin 1.7 converts UTF-16 Windows pathnames to/from multibyte
     * strings based on the values of LC_ALL, LC_CTYPE, and LANG when
     * the process started, regardless of later calls to setlocale().
     *
     * Therefore we must restart this process in order to ensure that
     * Windows pathname translation also uses UTF-8 in the narrow case.
     */
    execv(target, (char **)argv);

    perror("execv");
    exit(1);

#else /* not Cygwin 1.7+ */

    /* Environment modified, so try again.
     */
    if (! setlocale(LC_ALL, "")) {
        /* Failed to ensure UTF-8.
         */
        perror("setlocale");
        exit(1);
    }

#endif /* Cygwin 1.7+? */
#endif /* Cygwin 1.5? */

unset:

#if defined(__CYGWIN__) && CYGWIN_VERSION_DLL_MAJOR >= 1007

    /* Clear out the environment variable that we use to prohibit
     * multiple self-executions; it is an implementation detail
     * and should remain invisible.
     */
    if (unsetenv(VRPSX_LOADER_RECURSIVE)) {
        perror("unsetenv");
        exit(1);
    }

#endif /* defined(__CYGWIN__) && CYGWIN_VERSION_DLL_MAJOR >= 1007 */

    return;
}


/* Implements trace_expand except for allocation and termination.
 *
 * The expansion rules are as follows:
 *
 *     "@@" -> "@"
 *
 *     "@PID@" -> the process ID
 *
 *     "@NODE@" -> the hostname, typically as would be reported by "uname -n"
 *
 */
static size_t trace_expand_to(
    char const *path_template,
    char *target,
    size_t capacity)
{
    char const *sp = path_template;
    size_t index = 0;

#define TRACE_PUTC(CH)  {                       \
    char ch_2ef9570f = (CH);                    \
                                                \
    if (index < capacity) {                     \
        target[index] = ch_2ef9570f;            \
    }                                           \
                                                \
    ++index;                                    \
}

#define TRACE_PUTS(BEGIN)  {                                            \
    char const *begin_2ef9570f = (BEGIN);                               \
    size_t size_2ef9570f = strlen(begin_2ef9570f);                      \
    size_t prefix_2ef9570f = capacity > index ? capacity - index : 0;   \
                                                                        \
    if (prefix_2ef9570f > size_2ef9570f) {                              \
        prefix_2ef9570f = size_2ef9570f;                                \
    }                                                                   \
                                                                        \
    memcpy(target + index, begin_2ef9570f, prefix_2ef9570f);            \
    index += size_2ef9570f;                                             \
}

    while (*sp) {
        if (*sp != '@') {
            TRACE_PUTC(*sp++);
        } else {
            char const *name_begin = sp + 1;
            char const *name_end = name_begin;

            while (*name_end != '@' && *name_end) {
                ++name_end;
            }

            switch (name_end - name_begin) {
                case 0:
                    // "@@" -> "@"

                    TRACE_PUTC('@');
                    sp = name_end;
                    break;

                case 3:
                    if (0 == memcmp(name_begin, "PID", 3)) {
                        // "@PID@" -> the process ID

                        char buf[(sizeof(pid_t)*CHAR_BIT+2)/3+2];

                        sprintf(buf, "%d", getpid());

                        TRACE_PUTS(buf);
                        sp = name_end;
                    }
                    break;

                case 4:
                    if (0 == memcmp(name_begin, "NODE", 4)) {
                        // "@NODE@" -> the hostname, typically
                        // as would be reported by "uname -n"

                        struct utsname un;
                        int ern = uname(&un);
                        char const *node;

                        if (ern) {
                            node = "unknown";
                        } else {
                            node = un.nodename;
                        }

                        TRACE_PUTS(node);
                        sp = name_end;
                    }
                    break;
            }

            if (sp != name_end) {
                TRACE_PUTC(*sp++);
            } else {
                sp += !!*sp;
            }
        }
    }

#undef TRACE_PUTS
#undef TRACE_PUTC

    return index;
}

/* Allocates and returns the result of expanding the given path template.
 * The caller is responsible for deallocating the returned pointer with "free".
 * The expansion rules are in the function header comments for trace_expand_to.
 */
static char *trace_expand(
    char const *path_template)
{
    size_t size = trace_expand_to(path_template, NULL, 0);
    char *expansion = malloc(size + 1);
    size = trace_expand_to(path_template, expansion, size);
    expansion[size] = '\0';
    return expansion;
}


#define OPERATIONS_INITIALIZATION(PREFIX)                       \
                                                                \
static operations_t const VRPSX_CAT2(PREFIX, operations) = {    \
    &VRPSX_CAT2(PREFIX, get_library_path),                      \
    &VRPSX_CAT2(PREFIX, dlopen),                                \
    &VRPSX_CAT2(PREFIX, dlsym),                                 \
    &VRPSX_CAT2(PREFIX, synchronize_environment),               \
    &operations_end                                             \
};

OPERATIONS_INITIALIZATION();
OPERATIONS_INITIALIZATION(trace_);

static operations_t const *op = &operations;


/* Provides a way for prepare() to return a function
 * that simply returns a value it already computed.
 */
static int stub_result = 0;
static int stub(void) { return stub_result; }


/* We isolate as much as is convenient of the preparatory work
 * in a subroutine of main(), in order to minimize stack usage by
 * main() and thus provide more stack to the identified function.
 */
static invoker *prepare(void)
{
    static char const identified_functions_symbol[]
            = VRPSX_IDENTIFIED_FUNCTION_LIST_STR;
    static char const main_str[] = "main";

    char *trace_var = NULL;
    char *dlsym_error = NULL;
    char *library_path = NULL;
    void *library_handle = NULL;
    vrpsx_identified_function_t const *identified_function_list = NULL;
    vrpsx_identified_function_t const *identified = NULL;
    vrpsx_interface_identifier_t const *callable = NULL;
    vrpsx_embedding_t *embedding = NULL;
    invoker *(*constructor)(void (*)(void), vrpsx_embedding_t **) = NULL;
    invoker *call = NULL;
    trace_stack_t top;
    vrpsx_errno_t ern;

    establish_locale(arguments.array);

    /* The supported version of vrpsx_system_t requires this timezone setting:
     */
    tzset();

    trace_var = getenv("VRPSX_LOADER_TRACE_FILE");
    if (trace_var && *trace_var) {
        char *trace_expanded = trace_expand(trace_var);
        FILE *trace_file = fopen(trace_expanded, "w");

        if (trace_file) {
            char const *const *arg;

            direct_trace(trace_file);
            op = &trace_operations;

            trace_enter_begin(&top, main_str);
            for (arg = arguments.array; *arg; ++arg) {
                trace_string(&top, *arg);
            }
            trace_end(&top);
        } else {
            perror(trace_expanded);
        }

        free(trace_expanded);
    }

    library_path = op->get_library_path(arguments.array[0]);

    if (! library_path) {
        exit(1);
    }

    /* We never dlclose this library, because we will always have at
     * least one of the following reasons either to keep the library
     * loaded or to not bother to unload it explicitly:
     *
     *   1. We are about to terminate anyway.
     *
     *   2. We are running the identified function inside the library.
     *
     *   3. The identified function has returned but it may have
     *   spawned threads that are still running inside the library.
     */
    library_handle = op->dlopen(library_path, RTLD_LAZY | RTLD_GLOBAL);
    if (! library_handle) {
#if defined(__linux__)
        fprintf(stderr, "dlopen: %s\n", dlerror());
#else
        fprintf(stderr, "dlopen: %s: %s\n", library_path, dlerror());
#endif
        exit(1);
    }

    identified_function_list
            = op->dlsym(library_handle, identified_functions_symbol);
    if (! identified_function_list) {
        dlsym_error = dlerror();
#if defined(__CYGWIN__)
        if (! strcmp(dlsym_error, "No error")) {
            dlsym_error = "undefined symbol";
        }
#endif
        fprintf(stderr, "wrong vrpsx_identified_function array version: dlsym:"
#if defined(__CYGWIN__)
                /* Include library and symbol--Cygwin does not mention them. */
                " %s: %s: %s\n",
                library_path, dlsym_error, identified_functions_symbol
#else
                /* Linux already mentions the library and symbol. */
                " %s\n",
                dlsym_error
#endif
            );
        exit(1);
    }

    for (identified = identified_function_list;
         identified->function; ++identified) {
        callable = loader_main_supports(identified->interface_identifier);
        if (callable) {
            break;
        }
    }

    if (! callable) {
        for (identified = identified_function_list;
             identified->function; ++identified) {
            fprintf(
                stderr, "%s: %s: unsupported interface: %08lx ",
                arguments.array[0],
                library_path,
                (unsigned long)identified->interface_identifier);
            print_string(stderr, identified->display_name);
            fputc('\n', stderr);
        }
        exit(2);
    }

    free(library_path);
    library_path = NULL;

    ern = direct_embedding_acquire(&embedding);
    if (ern) {
        fprintf(
            stderr, "%s: error creating system interface: %s\n",
            arguments.array[0], strerror(ern));
        exit(1);
    }

    switch (*callable) {
        case vrpsx_system_main_IDENTIFIER:
            constructor = &system_preparation_construct;
            break;

        case vrpsx_embedding_main_IDENTIFIER:
            constructor = &embedding_preparation_construct;
            break;

        default:
            abort(); /* extend switch to handle supported interface */
            break;
    }

    call = constructor(identified->function, &embedding);

    /* Avoid interleaving our buffered output with any unbuffered
     * output that may be produced by the identified function.
     */
    fflush(stdout);
    fflush(stderr);

    /* If tracing is enabled, then we do not bother to try to optimize
     * stack space.  Instead, we directly call the desired functionality,
     * report the exit code, and then stub out the normal call mechanism.
     */
    if (op == &trace_operations) {
        trace_note_begin(&top, "client");
        trace_interface_identifier(&top, identified->interface_identifier);
        trace_end(&top);

        stub_result = call();

        trace_leave_begin(&top, main_str);
        trace_decimal_int(&top, stub_result);
        trace_end(&top);

        call = &stub;
    }

    return call;
}


/* We call "prepare" through a function pointer in order to prevent automatic
 * inlining, which would defeat the stack-conservation purpose of prepare().
 */
static invoker *(*prepare_ptr)(void) = &prepare;


int main(
    int argc,
    char **argv)
{
    VRPSX_ARGV_INITIALIZE_ARRAY(&arguments, argc, argv);
    return prepare_ptr()();
}
